/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory.createExplorationResult;
import static org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory.createExplorationSolution;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.CompositeExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.SolutionQuantification;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IExplorationSolution;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.ExplorationTransformationInputs;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.service.ITransformationService;
import org.opt4j.core.Genotype;
import org.opt4j.core.Individual;
import org.opt4j.core.Objective;
import org.opt4j.core.Objectives;
import org.opt4j.core.optimizer.Archive;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;

/**
 * Result of an Opt4J run, i.e. the {@link SolutionQuantification}s for all pareto-optimal
 * {@link IExplorationSolution}s.
 * 
 * @author barner
 */
public class Opt4JExplorationSolutionSet {

	/** Associates {@link ExplorationTarget}s with their Opt4J {@link Objective} equivalents. */
	private Map<ExplorationTarget<?>, Objective> targetObjectiveAssocs;

	/** References the transformed {@link ExplorationSolution}. */
	private ExplorationSolution explorationSolution;

	/**
	 * Constructs an {@link CompositeExplorationSolution} representation from the results obtained
	 * by an Opt4J
	 * run, i.e. the {@link SolutionQuantification}s for all pareto-optimal
	 * {@link IExplorationSolution}s.
	 * 
	 * @throws Exception
	 *             if the evaluation result for an {@link Objective} of an individual cannot be
	 *             identified.
	 */
	public Opt4JExplorationSolutionSet(Archive archive,
			Map<ExplorationTarget<?>, Objective> targetObjectiveAssocs) throws Exception {
		this.targetObjectiveAssocs = targetObjectiveAssocs;
		explorationSolution = createExplorationSolution();

		for(Individual individual : archive) {
			SingleExplorationSolution solution =
					createSingleExplorationSolution(individual.getObjectives());
			explorationSolution.getSolutions().add(solution);
			// TODO: Not type-safe. Opt4J has to be modified here.
			addOutputSuperSets(solution, ((ProvidingPhenotype)individual.getPhenotype()));
		}
	}

	/** Returns the transformation result. */
	public ExplorationSolution getExplorationSolution() {
		return explorationSolution;
	}

	/** Transforms {@link Objectives} into a {@link SingleExplorationSolution}. */
	private <T> SingleExplorationSolution createSingleExplorationSolution(Objectives objectives)
			throws ExplorationException {
		SingleExplorationSolution solution =
				DesignSpaceExplorationModelElementFactory.createSingleExplorationSolution();
		for(Entry<ExplorationTarget<?>, Objective> targetAssoc : targetObjectiveAssocs.entrySet()) {
			// Determine the ExplorationTarget class to identify the correct return value type.
			Objective objective = targetAssoc.getValue();
			@SuppressWarnings("unchecked") ExplorationTarget<T> explorationTarget =
					(ExplorationTarget<T>)targetAssoc.getKey();
			Class<?> expectedResultType = explorationTarget.getResultType();
			if(expectedResultType == null) {
				throw new ExplorationException(
						"Could not determine the result type of the exploration target " +
								explorationTarget.getName() +
								". The ExplorationTarget was wrongly constructed.");
			}

			// Type check the result vs. the expected type defined by the exploration target.
			Object result = objectives.get(objective).getValue();
			if(result != null && !result.getClass().equals(expectedResultType)) {
				// Note: if the result is null, a serious constraint violation is indicated. Its
				// corresponding solution will be filtered out in a later stage.
				throw new IllegalArgumentException(
						"Incompatible evaluation result type for the exploration target " +
								explorationTarget.getName() + ".");
			}
			@SuppressWarnings("unchecked") T castedResult = (T)result;

			ExplorationResult<T> explorationResult = createExplorationResult();
			explorationResult.setResult(castedResult);
			solution.put(explorationTarget, explorationResult);
		}

		return solution;
	}

	/**
	 * Adds the producable {@link SuperSet}s as a {@link SuperSetMap} to the given
	 * {@link SingleExplorationSolution}. It delegates the transformation operation to the
	 * {@link ITransformationService}.
	 * 
	 * @param solution
	 *            where the transformed {@link SuperSet}s are added.
	 * @param phenotype
	 *            output solution produced by the DSE.
	 * @return Constructed {@link SuperSetMap} from the exploration solutions.
	 * @throws Exception
	 *             if the transformation servies fails.
	 */
	@SuppressWarnings("unchecked")
	private SuperSetMap addOutputSuperSets(SingleExplorationSolution solution,
			ProvidingPhenotype phenotype) throws Exception {
		// FIXME: HACK: Currently, only the Deployment exploration is used, whereas the solution
		// transformation framework is non-functional at the moment.
		Collection<Class<?>> outputTypes = new ArrayList<>();
		// In a later stage, we would pass a list of generated "generatable" SuperSets derived
		// from the Phenotypes. This is the inverse setup of the model-oriented transformations
		// where we request desired types. Here, we'd calculate all types that can be produced.
		outputTypes.add(SuperSetMap.class);

		ClassToInstanceMap<Phenotype> innerSolutionMap = MutableClassToInstanceMap.create();
		// Hybrid Genotypes must be added first such that they can be overridden by Phenotypes.
		for(Genotype genotype : phenotype.getGenotypeMap().values()) {
			if(genotype instanceof Phenotype) {
				innerSolutionMap.put((Class<? extends Phenotype>)genotype.getClass(),
						(Phenotype)genotype);
			}
		}
		innerSolutionMap.putAll(phenotype.getPhenotypeMap());

		ExplorationTransformationInputs inputs =
				new ExplorationTransformationInputs(null, innerSolutionMap);
		Map<Class<IModelElement>, IModelElement> solutionModels =
				getService(IExplorationTransformationService.class).getTransformedModels(
						IExplorationTransformationService.TransformationContext.DSE_INTERNAL, null,
						inputs, outputTypes, "");
		SuperSetMap superSetMap = (SuperSetMap)solutionModels.get(SuperSetMap.class);

		// Add the transformed SuperSets to the passed ExplorationSolution.
		superSetMap.keySet().forEach(
				k -> solution.putSolutionModel((Class<IModelElement>)k,
						superSetMap.get((Class<IModelElement>)k)));
		return superSetMap;
	}
}
