/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import static java.util.Collections.emptyMap;
import static org.fortiss.af3.exploration.util.ExplorationUtils.isDebugVerboseEnabled;
import static org.fortiss.af3.schedule.utils.MathUtils.lcmDoubleCollection;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import javafx.concurrent.Task;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.AcyclicTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.SignalAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.IPlatformCommunicationResource;
import org.fortiss.af3.platform.model.IPlatformExport;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.jgrapht.Graphs;
import org.jgrapht.alg.connectivity.ConnectivityInspector;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Adapter to the system model used as input for the DSE which is created for a specific DSE run.
 * 
 * @author barner, huang
 */
public class SystemModelAdapter<C, E, SO, SI, G, TR, M> extends AbstractModule {

	/**
	 * Constant defining the allowed tolerance for the calculation of the greatest common divisor
	 * used in the calculation of the effective periods.
	 */
	protected final static int GCD_PRECISION = 3;

	/** Logical architecture of this {@link SystemModelAdapter}. */
	protected ILogicalArchitectureAdapter<?> logicalArchitecture;

	/** Platform architecture of this {@link SystemModelAdapter}. */
	protected IPlatformArchitectureAdapter<?> platformArchitecture;

	/** Deployable software components, to be mapped to deployment targets provided by platform. */
	protected Collection<ITaskAdapter<C>> tasks;

	/** Collection of deployable components from the base model. */
	protected Collection<C> taskObjects;

	/** Collection of signals exchanged between tasks. */
	protected Collection<ISignalAdapter<SO, SI>> signals;

	/**
	 * Look up table to determine the input/output directions of {@link IConnector} types of the
	 * target {@link PlatformArchitecture}.
	 */
	protected Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT = emptyMap();

	/**
	 * Deployable software components that may be replicated (per default: all @{code
	 * deployableComponent}s.
	 */
	protected Collection<ITaskAdapter<C>> nonSafetyFunctionTasks;

	/**
	 * Collection of software components that will be actually executed. This includes realizations
	 * of "abstract" software components (-> interfaces) that are contained in the
	 * {@code deployableComponents} collection.
	 */
	protected Collection<ITaskAdapter<C>> abstractTasks;

	/** Mapper that associates Tasks with their {@link ITaskAdapter} abstractions. */
	protected TaskMapper<C> taskMapper;

	/**
	 * Deployment targets, i.e. the execution units of the target platform, where deployable
	 * software components will be mapped to.
	 */
	protected Collection<IExecutionUnitAdapter<E>> deploymentTargets;

	/** All {@link ExecutionUnit}s that are present in the target platform. */
	protected Collection<IExecutionUnitAdapter<E>> allPresentExecutionUnits;

	/** {@link Collection} of memories present in the system. */
	protected Collection<IMemoryUnitAdapter<M>> memoryUnits;

	/**
	 * Gateway units that are employed to manage the communication between multiple networks
	 * (connecting the same or different layers).
	 */
	protected Collection<IGatewayUnitAdapter<G>> gatewayUnits;

	// TODO Consider more than one transmission unit
	/** TDMA arbitrated transmission unit */
	protected Collection<ITransmissionUnitAdapter<TR>> transmissionUnits;

	/** Contains all available communication resources of the platform. */
	protected Collection<ICommunicationResourceAdapter<TR>> availableCommunicationResource;

	/** Contains all resources that are available in the given platform. */
	protected Collection<IResourceAdapter<?>> availableResources;

	/**
	 * Contains all deployment target resources, i.e. the deployment targets and their
	 * interconnects.
	 */
	protected Collection<IResourceAdapter<?>> deploymentTargetResources;

	/** Contains the connections between the {@link IResourceAdapter}s of the platform. */
	protected Collection<IResourceConnectionAdapter> platformConnections;

	/**
	 * Collection of the safety functions adapters that represent the safety functions from the
	 * input models.
	 */
	protected Collection<ISafetyFunctionAdapter<C>> safetyFunctionAdapters = new ArrayList<>();

	/** The {@link PlatformCommunicationGraphEncoding} that is given by the current platform. */
	protected PlatformCommunicationGraphEncoding platformCommunicationGraph;

	/** The set of independent {@link DefaultTaskGraph}s derived from the logical architecture. */
	protected Set<TaskGraph> taskGraphs;

	/**
	 * The set of independent directed acyclic {@link DefaultTaskGraph}s derived from the logical
	 * architecture.
	 */
	protected Collection<AcyclicTaskGraph> acyclicTaskGraphs;

	/** The "messages" emitted by the sender {@link ITaskAdapter}. */
	protected Multimap<ITaskAdapter<C>, ISignalAdapter<SO, ?>> emittedMessagesBySender =
			HashMultimap.create();

	/** Hyper-period of independent tasks graphs */
	protected Double hyperPeriod;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link DefaultTaskGraph}.
	 */
	protected Map<ITaskAdapter<C>, TaskGraph> tasksInGraph;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link AcyclicTaskGraph}.
	 */
	protected Map<ITaskAdapter<C>, AcyclicTaskGraph> componentInAcyclicGraph;

	/**
	 * Mapping virtual resource -> hardware resource onto which virtual resource is (statically)
	 * mapped, e.g., partition -> hardware execution unit / processor core.
	 * 
	 * The virtual resources are considered as deployment targets where {@link IRequestAdapter}s can
	 * be allocated to.
	 * 
	 * However, deployment targets running on the same core share the same time line.
	 * Hence,a map between the deployment targets and the "time line" is maintained which
	 * will be considered during scheduling to separate partitions in time.
	 */
	protected Map<IResourceAdapter<?>, IResourceAdapter<?>> virtualResourceToHardwareResourceMap;

	/** Defines the fault containment region defined for the target {@link PlatformArchitecture}. */
	protected Class<? extends IPlatformResource> faultContainmentRegion;

	// TODO: Make the task graph setup independent from the AF3-Component Architecture and write a
	// wrapper in the specific AF3 SystemModelAdapter.
	/**
	 * Sets up the {@link DefaultTaskGraph}s for all (independent) connected sub-component
	 * model contained in the logical {@link ComponentArchitecture}. This method
	 * also determines the hyper-period of all {@link DefaultTaskGraph}s.
	 * 
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	protected void setupTaskGraphs() throws Exception {
		// Create Mappers that allow to trace an input object to its internal adapter.
		taskMapper = createTaskMapper();

		// - Add data dependencies to {@link TaskGraph}s
		// - Print some debugging output
		// - Determine the hyper-period of all {@link TaskGraph}s contained in this {@link
		// SystemModelAdapter}.
		taskGraphs = new HashSet<TaskGraph>();
		acyclicTaskGraphs = new ArrayList<>();

		// Temporary task graph: add vertices (= software components / tasks that need to be mapped
		// to processing elements, and that potentially require data from other software components)
		DefaultTaskGraph<?> temporaryTaskGraph = new DefaultTaskGraph<>();
		// TODO: Use generics in the task graph
		temporaryTaskGraph.addAllTask((Collection<ITaskAdapter<?>>)(Collection<?>)tasks);
		temporaryTaskGraph.addAllSignals((Collection<ISignalAdapter<?, ?>>)(Collection<?>)signals);

		// Separate temporary task graph into multiple TaskGraphs that consist of exactly one
		// connected graph component. I.e., there are no data dependencies between the resulting new
		// TaskGraphs.
		// Create a generic TaskGraph of the logical architecture (can contain cycles)
		taskGraphs = createDefaultTaskGraphs(temporaryTaskGraph);
		tasksInGraph = registerTaskGraphsForTasks(taskGraphs);

		// Construct an Directed Acyclic TaskGraph (e.g., utilized in scheduling).
		acyclicTaskGraphs =
				taskGraphs.stream().map(tg -> new AcyclicTaskGraph(tg))
						.collect(Collectors.toList());
		componentInAcyclicGraph = registerTaskGraphsForTasks(acyclicTaskGraphs);

		// Print some debugging output: print the taskgraphs and display them in a dialog.
		if(isDebugVerboseEnabled()) {
			printAndDisplayTaskGraphs(taskGraphs);
			printAndDisplayTaskGraphs(acyclicTaskGraphs);
		}

		Collection<TaskGraph> allTaskGraphs = new ArrayList<TaskGraph>();
		allTaskGraphs.addAll(taskGraphs);
		allTaskGraphs.addAll(acyclicTaskGraphs);
		hyperPeriod = determineHyperPeriod(allTaskGraphs);
	}

	/**
	 * Creates a set of independent {@link TaskGraph}s of the given type ({#link graphClass}). The
	 * subgraphs may have different implementation to construct themselves (via
	 * addDataDependencies).
	 * 
	 * @throws InvalidTimingModelException
	 *             if the periods for a single task graph cannot be set up.
	 */
	private Set<TaskGraph> createDefaultTaskGraphs(TaskGraph monolithicTaskGraph)
			throws InvalidTimingModelException {
		Set<TaskGraph> independentTaskGraphs = new HashSet<>();
		// Compute vertex sets of maximum connected components of temporary task graph
		ConnectivityInspector<ITaskAdapter<?>, SignalAdapterWeightedEdge> ci =
				new ConnectivityInspector<ITaskAdapter<?>, SignalAdapterWeightedEdge>(
						monolithicTaskGraph.getGraph());
		List<Set<ITaskAdapter<?>>> connectedSets = ci.connectedSets();
		// Reconstruct TaskGraphs from vertex sets of connected components
		for(Set<ITaskAdapter<?>> connectedSet : connectedSets) {
			TaskGraph taskGraph = new DefaultTaskGraph<>();
			Graphs.addAllVertices(taskGraph.getGraph(), connectedSet);
			for(ISignalAdapter<?, ?> signal : signals) {
				if(connectedSet.contains(signal.getSource())) {
					taskGraph.addSignal(signal);
				}
			}
			// Check if tasks within current sub TaskGraph share the same common period
			taskGraph.updatePeriod();
			independentTaskGraphs.add(taskGraph);
		}
		return independentTaskGraphs;
	}

	/**
	 * Creates the {@link TaskMapper} data structure from the input {@link Task}s and the
	 * {@link ITaskAdapter} abstractions .
	 */
	private TaskMapper<C> createTaskMapper() {
		Map<C, ITaskAdapter<C>> taskMap = new HashMap<>();
		for(ITaskAdapter<C> task : tasks) {
			taskMap.put(task.getObject(), task);
		}
		return new TaskMapper<C>(taskMap);
	}

	/**
	 * Adds the given collection of {@link ISafetyFunctionAdapter}s to {@code this}
	 * {@link SystemModelAdapter} that represents the input system models. In contrast to other
	 * input parameters, the {@link ISafetyFunctionAdapter}s are created dynamically if a
	 * safety constraint is present which results in the safety functions to be
	 * extracted <b>after</b> the {@link SystemModelAdapter} has been constructed.
	 * <p>
	 * This method updates the set of replicable components by removing those components that are
	 * referenced by the safety function.
	 */
	// TODO: remove the enclosed cast. The difficulty is that the ISafetyFunctionAdapters are added
	// from within the DSE where the system types are not known any more.
	// It would be more appropriate to shift this addition before launching the DSE.
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void
			addSafetyFunctionAdapters(Collection<ISafetyFunctionAdapter> safetyFunctionAdapters) {
		if(safetyFunctionAdapters == null) {
			return;
		}
		this.safetyFunctionAdapters
				.addAll((Collection<? extends ISafetyFunctionAdapter<C>>)safetyFunctionAdapters);

		for(ISafetyFunctionAdapter<?> sfAdapter : safetyFunctionAdapters) {
			nonSafetyFunctionTasks.removeAll(sfAdapter.getDiagnosticComponents());
			nonSafetyFunctionTasks.removeAll(sfAdapter.getRealizingComponents());
			nonSafetyFunctionTasks.remove(sfAdapter.getOutputComponent());
		}
	}

	/**
	 * Creates a map for the given collection of {@link TaskGraph}s which allows to identify the
	 * {@link TaskGraph} that contains the given {@link ITaskAdapter} as a vertex.
	 */
	// TODO: Here, we know that the IDeployableCOmponentAdapters are constructed using Components.
	// However, we should try to avoid the "SuppressWarnings".
	@SuppressWarnings("unchecked")
	private <T extends TaskGraph> Map<ITaskAdapter<C>, T> registerTaskGraphsForTasks(
			Collection<T> graphs) {
		Map<ITaskAdapter<C>, T> componentToGraphMap = new HashMap<ITaskAdapter<C>, T>();

		for(T taskGraph : graphs) {
			for(ITaskAdapter<?> deployableComponent : taskGraph.getGraph().vertexSet()) {
				componentToGraphMap.put((ITaskAdapter<C>)deployableComponent, taskGraph);
			}
		}

		return componentToGraphMap;
	}

	/**
	 * Calculates and returns the hyperperiod of the set of all given {@link TaskGraph}s by
	 * calculating the least common multiple from the hyperperiods of the {@link TaskGraph}s.
	 */
	private Double determineHyperPeriod(Collection<TaskGraph> allTaskGraphs) {
		Set<Double> taskgraphPeriods = new TreeSet<Double>();
		for(TaskGraph currentGraph : allTaskGraphs) {
			Double period = currentGraph.getPeriod();
			if(period != null && period > 0) {
				taskgraphPeriods.add(period);
			}
		}

		if(taskgraphPeriods.isEmpty()) {
			return null;
		}
		return lcmDoubleCollection(taskgraphPeriods, GCD_PRECISION);
	}

	/**
	 * Prints the given collection of {@link TaskGraph}s to the standard output and displays them in
	 * a separate dialog.
	 */
	private <T extends TaskGraph> void printAndDisplayTaskGraphs(Collection<T> taskGraphs) {
		for(T taskGraph : taskGraphs) {
			System.out.println("Application " + taskGraph + ":" + " period = " +
					taskGraph.getPeriod());
			System.out.println(taskGraph);
			System.out.println("");

			// Open a dialog for the visualization of the TaskGraph
			taskGraph.display();
		}
	}

	/**
	 * Setups the {#link emittedMessagesBySender} map which makes it possible to easily identify the
	 * {@link ISignalAdapter}s that are emitted by the given {@link ITaskAdapter}.
	 */
	protected void setupEmittedMessages() {
		for(ITaskAdapter<C> currentSender : tasks) {
			for(ISignalAdapter<SO, ?> sig : signals) {
				if(currentSender == sig.getSource()) {
					emittedMessagesBySender.put(currentSender, sig);
				}
			}
		}
	}

	/** Returns the logical architecture that is associated with this {@link SystemModelAdapter}. */
	public ILogicalArchitectureAdapter<?> getLogicalArchitecture() {
		return logicalArchitecture;
	}

	/**
	 * Returns the platform architecture (Middleware/HW) that is associated with this
	 * {@link SystemModelAdapter}.
	 */
	public IPlatformArchitectureAdapter<?> getPlatformArchitecture() {
		return platformArchitecture;
	}

	/**
	 * Returns the deployable (software) component from the original logical architecture. May
	 * contain "abstract" {@link Component}s.
	 */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<ITaskAdapter<?>> getDeployableComponents() {
		return (Collection<ITaskAdapter<?>>)(Collection<?>)tasks;
	}

	/**
	 * Returns the deployable (software) component from the original logical architecture that my be
	 * replicated. May contain "abstract" {@link Component}s.
	 */
	@Provides
	@Named("NonSafetyFunctionTasks")
	public Collection<ITaskAdapter<C>> getNonSafetyFunctionTasks() {
		return nonSafetyFunctionTasks;
	}

	/**
	 * Returns the deployable (software) components that can be actually executed on
	 * {@link ExecutionUnit}s, i.e. software components that are backed by an implementation.
	 */
	public Collection<ITaskAdapter<C>> getAbstractTasks() {
		return abstractTasks;
	}

	/** Returns the mapper tracing tasks to its adapter. */
	@SuppressWarnings("unchecked")
	@Provides
	public TaskMapper<EObject> getTaskMapper() {
		// We need to use EObject for now since Guice requires fully specified types.
		return (TaskMapper<EObject>)taskMapper;
	}

	/**
	 * Returns the {@link ITaskAdapter} that wraps the given {@link EObject}. Here,
	 * only those {@link ITaskAdapter} are searched which can be actually executed on
	 * {@link ExecutionUnit}s, i.e. "abstract" components are not considered.
	 */
	public synchronized ITaskAdapter<C>
			getRealizeableDeployableComponentOf(EObject originalEObject) {
		for(ITaskAdapter<C> deployableComponentAdapter : abstractTasks) {
			if(deployableComponentAdapter.getObject() == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link ISignalAdapter} corresponding the sender and receiver {@link Component}.
	 * Returns null if no channel exists between the two given components.
	 */
	public synchronized ISignalAdapter<SO, ?> getSignalAdapter(ITaskAdapter<C> source,
			ITaskAdapter<C> target) {
		for(ISignalAdapter<SO, ?> sig : signals) {
			if(sig.getSource() == source && sig.getTargets().contains(target)) {
				return sig;
			}
		}
		return null;
	}

	/**
	 * Returns the deployment targets, i.e. the execution units of the target platform,
	 * where deployable software components will be mapped to.
	 */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<IExecutionUnitAdapter<?>> getDeploymentTargets() {
		return (Collection<IExecutionUnitAdapter<?>>)(Collection<?>)deploymentTargets;
	}

	/** Returns the {@link IMemoryUnitAdapter} present in the target platform. */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<IMemoryUnitAdapter<?>> getMemoryUnits() {
		return (Collection<IMemoryUnitAdapter<?>>)(Collection<?>)memoryUnits;
	}

	/** Returns the gateway units present in the platform. */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<IGatewayUnitAdapter<?>> getGatewayUnits() {
		return (Collection<IGatewayUnitAdapter<?>>)(Collection<?>)gatewayUnits;
	}

	/** Returns the systems TDMA arbitrated transmission unit */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<ITransmissionUnitAdapter<?>> getTransmissionUnits() {
		return (Collection<ITransmissionUnitAdapter<?>>)(Collection<?>)transmissionUnits;
	}

	/** Returns all the {@link IResourceAdapter}s present in the platform. */
	@Provides
	public Collection<IResourceAdapter<?>> getAvailableResources() {
		return availableResources;
	}

	/**
	 * Returns the hardware resource (e.g., processor core) onto which a virtual resource
	 * (e.g. partition) is (statically) mapped.
	 */
	public IResourceAdapter<?> getHardwareResource(IResourceAdapter<?> virtualResource) {
		return virtualResourceToHardwareResourceMap.get(virtualResource);
	}

	/** Returns the present connections between platform resources. */
	@Provides
	public Collection<IResourceConnectionAdapter> getResouceConnections() {
		return platformConnections;
	}

	/**
	 * Returns the {@link ISafetyFunctionAdapter}s constructed from the safety functions that are
	 * present in the input models.
	 */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<ISafetyFunctionAdapter<?>> getSafetyFunctionAdapters() {
		if(safetyFunctionAdapters != null) {
			return (Collection<ISafetyFunctionAdapter<?>>)(Collection<?>)safetyFunctionAdapters;
		}
		return Collections.emptySet();
	}

	/**
	 * Returns the {@link PlatformCommunicationGraphEncoding} required for routing of the messages
	 * located
	 * in different parts of the platform.
	 */
	public PlatformCommunicationGraphEncoding getPlatformCommunicationGraph() {
		return platformCommunicationGraph;
	}

	/**
	 * Returns the {@link TaskGraph}s of the demanded type that has been constructed by {@code this}
	 * {@link SystemModelAdapter}, if any exists. Otherwise, an empty collection is returned.
	 */
	@SuppressWarnings("unchecked")
	public <T extends TaskGraph> Collection<T> getTaskGraphsByType(Class<T> tgType) {
		if(DefaultTaskGraph.class.isAssignableFrom(tgType)) {
			return (Collection<T>)getTaskGraphs();
		} else if(AcyclicTaskGraph.class.isAssignableFrom(tgType)) {
			return (Collection<T>)getAcyclicTaskGraphs();
		}
		return Collections.emptySet();
	}

	/**
	 * Returns all independent {@link DefaultTaskGraph}s contained in this
	 * {@link SystemModelAdapter}.
	 */
	@Provides
	@Named("OriginalTaskGraphs")
	public Set<TaskGraph> getTaskGraphs() {
		return taskGraphs;
	}

	/**
	 * Returns all independent {@link DefaultTaskGraph}s contained in this
	 * {@link SystemModelAdapter}.
	 */
	public Collection<? extends TaskGraph> getAcyclicTaskGraphs() {
		return acyclicTaskGraphs;
	}

	/**
	 * Returns the {@link DefaultTaskGraph} in which a given deployable component (=task) is
	 * contained.
	 */
	public TaskGraph getTaskGraph(ITaskAdapter<?> deployableComponent) {
		return tasksInGraph.get(deployableComponent);
	}

	/**
	 * Returns the {@link DefaultTaskGraph} in which a given deployable component (=task) is
	 * contained.
	 */
	public DefaultTaskGraph<?> getAcyclicTaskGraph(ITaskAdapter<?> deployableComponent) {
		return componentInAcyclicGraph.get(deployableComponent);
	}

	/**
	 * Returns the {@link ISignalAdapter}s "emitted" by the given {@link ITaskAdapter}. Returns an
	 * empty collection if the given {@link ITaskAdapter} does not emit anything.
	 */
	public Collection<ISignalAdapter<SO, ?>> getEmmittedMessages(ITaskAdapter<?> sender) {
		// TODO: use another representation to make an access via the reference obsolete
		if(sender.isInstantiation()) {
			sender = sender.getReplacedComponent();
		}
		@SuppressWarnings("unchecked") ITaskAdapter<C> origSender =
				taskMapper.getAdapter((C)sender.getObject());
		return emittedMessagesBySender.get(origSender);
	}

	/**
	 * Returns the hyper period of all tasks in the system, i.e. the hyper-period of all independent
	 * {@link DefaultTaskGraph}s.
	 */
	@Provides
	@Named("HyperPeriod")
	public Double getHyperPeriod() {
		return hyperPeriod;
	}

	/** Returns whether the given {@link Component} is a deployable {@link Component}. */
	public boolean isDeployableComponent(Component component) {
		return taskObjects.contains(component);
	}

	/** Returns the input output behavior of the given platform connector. */
	public ConnectionType getConnectionType(Class<? extends IConnector> connector) {
		return connectionTypeLUT.get(connector);
	}

	// ===============================================================
	// To be removed / generalized
	// ===============================================================

	/**
	 * <p>
	 * Determines the effective target {@link IConnector}s of a given "source" {@link IConnector}.
	 * Elements/{@link IConnector}s that are irrelevant to communication within the platform are not
	 * added to the target list. Instead, the connectors that are accessible by the source
	 * {@link IConnector} are traversed until relevant targets are identified, like
	 * {@link ExecutionUnit}s selected for the deployment.
	 * </p>
	 * <p>
	 * <b>NOTE:</b> This method may return targets that are not directly involved in the
	 * communication between {@link ExecutionUnit}s. However, 'hull' elements (or similar elements)
	 * that may "disturb" the communication between {@link ExecutionUnit}s will not appear.
	 * </p>
	 * <p>
	 * <b>Implementation Note:</b> The <i>valid</i> targets are determined via path walk following
	 * all connections from the source {@link IConnector} until <i>valid</i> targets are found.
	 * </p>
	 * 
	 * @throws Exception
	 *             if the references between the {@link IModelElement}s are incorrectly specified.
	 */
	protected Collection<IConnector> getEffectiveConnectionTargets(IConnector sourceConnector,
			Collection<ExecutionUnit> af3DeploymentTargets) throws Exception {
		Collection<IConnector> targetConnectors = new HashSet<IConnector>();
		if(sourceConnector == null) {
			return targetConnectors;
		}

		LinkedList<IConnector> possibleTargetConnectors = new LinkedList<IConnector>();
		Collection<IConnector> visitedTargetConnectors = new HashSet<IConnector>();

		possibleTargetConnectors.add(sourceConnector);
		while(!possibleTargetConnectors.isEmpty()) {
			IConnector nextTargetConnector = possibleTargetConnectors.poll();

			// This section handles the traversal between the elements of the same platform
			// architecture.
			targetConnectors.addAll(seekIntraPlatformTargetConnectors(sourceConnector,
					possibleTargetConnectors, visitedTargetConnectors, nextTargetConnector,
					af3DeploymentTargets));

			visitedTargetConnectors.add(nextTargetConnector);
		}

		return targetConnectors;
	}

	/**
	 * Identifies those {@link IConnector}s that are actual target {@link IConnector}s of data sent
	 * out by the given {@code sourceConnector}. This methos is needed as in a
	 * {@link PlatformArchitecture}, there exist {@link IConnector}s which are attached to container
	 * elements that are not backed by physical interface ports (These are represented by the
	 * contained elements).
	 * 
	 * @param sourceConnector
	 *            Source {@link IConnector} for which the target {@link IConnector}s need to be
	 *            identified.
	 * @param possibleTargetConnectors
	 *            List to which those {@link IConnector}s are added that are directly connected to
	 *            the {@code sourceConnector} but are no "communication targets" as described above
	 *            an need further traversal.
	 * @param visitedTargetConnectors
	 *            Set of already traversed connectors (to avoid loops).
	 * @param nextTargetConnector
	 *            Current {@link IConnector} whose target {@link IConnector}s need to be searched.
	 * @param deploymentTargets
	 *            Set of execution units present on the target platform.
	 * @return Collection of {@link IConnector}s that are actual target {@link IConnector}s of the
	 *         {@code sourceConnector}.
	 */
	protected Collection<IConnector> seekIntraPlatformTargetConnectors(IConnector sourceConnector,
			LinkedList<IConnector> possibleTargetConnectors,
			Collection<IConnector> visitedTargetConnectors, IConnector nextTargetConnector,
			Collection<ExecutionUnit> deploymentTargets) {
		Collection<IConnector> targetConnectors = new ArrayList<IConnector>();

		Collection<IConnection> attachedConnections = new ArrayList<IConnection>();
		attachedConnections.addAll(nextTargetConnector.getOutgoing());
		attachedConnections.addAll(nextTargetConnector.getIncoming());

		for(IConnection connToTarget : attachedConnections) {
			IConnector connectorTgt = null;
			IConnector connectorSrc = null;
			if(connToTarget.getTarget().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getTarget();
				connectorSrc = connToTarget.getSource();
			}
			if(connToTarget.getSource().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getSource();
				connectorSrc = connToTarget.getTarget();
			}
			if(connectorTgt == null || connectorSrc == null) {
				throw new RuntimeException(
						"Error: \"floating\" connection found during transformation for the DSE.");
			}

			IHierarchicElement tgtElement = connectorTgt.getOwner();
			// Do not traverse back (to same connector) & not to the same IHierarchicElement.
			if(!visitedTargetConnectors.contains(connectorTgt) &&
					tgtElement != sourceConnector.getOwner()) {
				if(tgtElement instanceof IPlatformArchitectureElement &&
						isValidConnectionTarget((IPlatformArchitectureElement)tgtElement,
								deploymentTargets)) {
					targetConnectors.add(connectorTgt);
				} else if(tgtElement == connectorSrc.getOwner().getContainer() &&
						!(nextTargetConnector instanceof IPlatformExport)) {
					// Treatment of hull ports: Do not walk a path that would just traverse a 'hull'
					// port. Hence, just consider the ports to the "outside" (upper level).
					for(IConnection currConn : connectorTgt.getOutgoing()) {
						IConnector currTgtConnector = currConn.getTarget();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}
					for(IConnection currConn : connectorTgt.getIncoming()) {
						IConnector currTgtConnector = currConn.getSource();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}

					visitedTargetConnectors.add(connectorTgt);
				} else {
					possibleTargetConnectors.add(connectorTgt);
				}
			}
		}

		return targetConnectors;
	}

	/**
	 * Returns whether the given {@link IHierarchicElement} is a valid connection end-point. These
	 * end-points are either {@link ExecutionUnit}s selected as deployment targets for
	 * {@link Component}s or {@link IPlatformCommunicationResource}s whose internal structure is not
	 * modeled.
	 */
	protected boolean isValidConnectionTarget(IPlatformArchitectureElement element,
			Collection<ExecutionUnit> deploymentTargets) {
		if(element instanceof IPlatformResource && deploymentTargets.contains(element)) {
			return true;
		} else if(element instanceof IPlatformCommunicationResource &&
				element.getContainedElements().isEmpty()) {
			return true;
		}
		return false;
	}

	/** Returns the fault containment region specified for the target platform. */
	public Class<? extends IPlatformResource> getFaultContainmentRegion() {
		return faultContainmentRegion;
	}
}
